/*
  Test for duplicate timer callbacks
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string>
#include <cstdlib>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/timers.h"
#include "esp_system.h"
#include "esp_task.h"
#include "esp_spi_flash.h"
#include "esp_log.h"
#include "driver/gpio.h"

#include "esp_wifi.h"
#include "lwip/err.h"
#include "lwip/sockets.h"
#include "lwip/sys.h"
#include "lwip/netdb.h"
#include "lwip/dns.h"

#if 0
#define LED_OUT GPIO_NUM_2
#endif

QueueHandle_t queue;
TimerHandle_t timer;

extern "C"
{
  void app_main(void);
  void start_wifi();
}


//
// CPU/MEM load: loop malloc, random memory pokes, free
//

void cpuload_task(void *pvParameters)
{
  const int loopms = 10000;
  const int memsize = 128*1024;
  uint8_t* mem;

  ESP_LOGI("cpuload_task", "running on core %d", xTaskGetAffinity(xTaskGetCurrentTaskHandle()));

  while (true)
  {
    mem = (uint8_t*) heap_caps_malloc(memsize, MALLOC_CAP_SPIRAM);
    if (!mem)
    {
      ESP_LOGE("cpuload_task", "unable to allocate %d bytes SPIRAM", memsize);
      vTaskDelay(5000 / portTICK_PERIOD_MS);
    }
    else
    {
      for (int i=0; i<loopms; i++)
        mem[rand() % memsize] = i % 256;
      vTaskDelay(1); // 10 ms yield
      free(mem);
    }
  }
}


//
// Just a dummy callback for the periodic timers
//

void dummy_callback(TimerHandle_t timer)
{
  static int led = 0;
  led = 1 - led;
#ifdef LED_OUT
  gpio_set_level(LED_OUT, led);
#endif
}


//
// The single shot timer callback: this occasionally gets called twice.
// To detect the duplicate calls we remember the timer ID locally and
// check for the new call carrying the same ID again.
//

void timer_callback(TimerHandle_t timer)
{
  static uint32_t last_tid = 0;
  uint32_t tid = (uint32_t) pvTimerGetTimerID(timer);

  if (tid == last_tid)
  {
    // this should never happen
    ESP_LOGE("timer_callback", "duplicate callback detected on tid=%u", tid);
    return;
  }
  last_tid = tid;

  // signal app_task to continue:
  while (xQueueSend(queue, &tid, 0) != pdTRUE);
}


//
// app_task: loop single shot timer runs with unique timer IDs,
//  synchronized to the timer by the queue
//

void app_task(void *userdata)
{
  uint32_t tid = 0;
  uint32_t msg;

  ESP_LOGI("app_task", "running on core %d", xTaskGetAffinity(xTaskGetCurrentTaskHandle()));

  while (true)
  {
    // start next timer run:
    tid++;
    if (tid % 1000 == 0)
      ESP_LOGI("app_task", "checkpoint tid=%u", tid);
    while (xTimerChangePeriod(timer, 1, portMAX_DELAY) != pdPASS);
    vTimerSetTimerID(timer, (void *) tid);
    while (xTimerStart(timer, portMAX_DELAY) != pdPASS);
    
    // wait for msg:
    if (xQueueReceive(queue, (void *) &msg, 100) == pdTRUE)
    {
      if (msg != tid)
        ESP_LOGE("app_task", "tid mismatch: received %u, expected %u", msg, tid);
    }
    else
    {
      ESP_LOGW("app_task", "queue rx timeout -- timer callback not issued!");
    }
  }
}


void app_main()
{
  // Print chip information
  esp_chip_info_t chip_info;
  esp_chip_info(&chip_info);
  printf("This is ESP32 chip with %d CPU cores, WiFi%s%s, ",
          chip_info.cores,
          (chip_info.features & CHIP_FEATURE_BT) ? "/BT" : "",
          (chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");
  printf("silicon revision %d, ", chip_info.revision);
  printf("%dMB %s flash\n", spi_flash_get_chip_size() / (1024 * 1024),
          (chip_info.features & CHIP_FEATURE_EMB_FLASH) ? "embedded" : "external");

#ifdef LED_OUT
  gpio_set_direction(LED_OUT, GPIO_MODE_OUTPUT);
#endif

  // Start wifi:
  start_wifi();
  vTaskDelay(5000 / portTICK_PERIOD_MS);

  // Start CPU/MEM load:
  xTaskCreate(&cpuload_task, "cpuload_task", 4096, NULL, 5, NULL);

  // create queue:
  queue = xQueueCreate(20, sizeof(uint32_t));

  // create some periodic timers:
  xTimerStart(xTimerCreate("per_timer", 5, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 6, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 7, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 10, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 25, pdTRUE, NULL, dummy_callback), portMAX_DELAY);

  // create single shot timer:
  timer = xTimerCreate("test_timer", 1, pdFALSE, NULL, timer_callback);

  // timer service is on core 0, start application task on core 1:
  xTaskCreatePinnedToCore(app_task, "app_task", 4096, NULL, 5, NULL, 1);
}
