/*
  test_memory_corruption.cpp
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string>
#include <cstdlib>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/timers.h"
#include "esp_system.h"
#include "esp_task.h"
#include "esp_spi_flash.h"
#include "esp_log.h"
#include "driver/gpio.h"

#include "esp_wifi.h"
#include "lwip/err.h"
#include "lwip/sockets.h"
#include "lwip/sys.h"
#include "lwip/netdb.h"
#include "lwip/dns.h"

#if 0
#define LED_OUT GPIO_NUM_2
#endif

QueueHandle_t queue;
TimerHandle_t timer;

extern "C"
{
  void app_main(void);
  void start_wifi(void);
}


// 
// HTTP request
// 

#define WEB_SERVER "192.168.2.100"
#define WEB_PORT 80
#define WEB_URL "http://192.168.2.100/f/ovms3.bin"

const char *REQUEST = "GET " WEB_URL " HTTP/1.0\r\n"
    "Host: " WEB_SERVER "\r\n"
    "User-Agent: esp-idf/1.0 esp32\r\n"
    "\r\n";

void http_get_task(void *pvParameters)
{
    const char *TAG = "http_get_task";

    struct addrinfo hints = {};
    struct addrinfo *res;
    struct in_addr *addr;
    int s, r, rlen_total;
    char recv_buf[64];

    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    while(1) {
        int err = getaddrinfo(WEB_SERVER, "80", &hints, &res);

        if(err != 0 || res == NULL) {
            ESP_LOGE(TAG, "DNS lookup failed err=%d res=%p", err, res);
            vTaskDelay(5000 / portTICK_PERIOD_MS);
            continue;
        }

        /* Code to print the resolved IP.
           Note: inet_ntoa is non-reentrant, look at ipaddr_ntoa_r for "real" code */
        addr = &((struct sockaddr_in *)res->ai_addr)->sin_addr;
        //ESP_LOGI(TAG, "DNS lookup succeeded. IP=%s", inet_ntoa(*addr));

        s = socket(res->ai_family, res->ai_socktype, 0);
        if(s < 0) {
            ESP_LOGE(TAG, "... Failed to allocate socket.");
            freeaddrinfo(res);
            vTaskDelay(1000 / portTICK_PERIOD_MS);
            continue;
        }
        //ESP_LOGI(TAG, "... allocated socket");

        if(connect(s, res->ai_addr, res->ai_addrlen) != 0) {
            ESP_LOGE(TAG, "... socket connect failed errno=%d", errno);
            close(s);
            freeaddrinfo(res);
            vTaskDelay(4000 / portTICK_PERIOD_MS);
            continue;
        }

        //ESP_LOGI(TAG, "... connected");
        freeaddrinfo(res);

        if (write(s, REQUEST, strlen(REQUEST)) < 0) {
            ESP_LOGE(TAG, "... socket send failed");
            close(s);
            vTaskDelay(4000 / portTICK_PERIOD_MS);
            continue;
        }
        //ESP_LOGI(TAG, "... socket send success");

        struct timeval receiving_timeout;
        receiving_timeout.tv_sec = 5;
        receiving_timeout.tv_usec = 0;
        if (setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, &receiving_timeout,
                sizeof(receiving_timeout)) < 0) {
            ESP_LOGE(TAG, "... failed to set socket receiving timeout");
            close(s);
            vTaskDelay(4000 / portTICK_PERIOD_MS);
            continue;
        }
        //ESP_LOGI(TAG, "... set socket receiving timeout success");

        /* Read HTTP response */
        rlen_total = 0;
        do {
            bzero(recv_buf, sizeof(recv_buf));
            r = read(s, recv_buf, sizeof(recv_buf)-1);
            if (r > 0) rlen_total += r;
            //for(int i = 0; i < r; i++) { putchar(recv_buf[i]); }
        } while(r > 0);

        close(s);

        ESP_LOGI(TAG, "read %d bytes", rlen_total);

        //vTaskDelay(1000 / portTICK_PERIOD_MS);
        //ESP_LOGI(TAG, "Starting again!");
    }
}


//
// CPU/MEM load
//

void cpuload_task(void *pvParameters)
{
  const int loopms = 10000;
  const int memsize = 128*1024;
  uint8_t* mem;

  ESP_LOGI("cpuload_task", "running on core %d", xTaskGetAffinity(xTaskGetCurrentTaskHandle()));

  while (true)
  {
    mem = (uint8_t*) heap_caps_malloc(memsize, MALLOC_CAP_SPIRAM);
    if (!mem)
    {
      ESP_LOGE("cpuload_task", "unable to allocate %d bytes SPIRAM", memsize);
      vTaskDelay(5000 / portTICK_PERIOD_MS);
    }
    else
    {
      for (int i=0; i<loopms; i++)
        mem[rand() % memsize] = i % 256;
      vTaskDelay(1); // 10 ms yield
      free(mem);
    }
  }
}


void dummy_callback(TimerHandle_t timer)
{
  static int led = 0;
  led = 1 - led;
#ifdef LED_OUT
  gpio_set_level(LED_OUT, led);
#endif
}


void timer_callback(TimerHandle_t timer)
{
  static uint32_t last_tid = 0;
  uint32_t tid = (uint32_t) pvTimerGetTimerID(timer);

  if (tid == last_tid)
  {
    ESP_LOGE("timer_callback", "duplicate callback detected on tid=%u", tid);
    return;
  }
  last_tid = tid;

  while (xQueueSend(queue, &tid, 0) != pdTRUE);
}


void app_task(void *userdata)
{
  uint32_t tid = 0;
  uint32_t msg;

  ESP_LOGI("app_task", "running on core %d", xTaskGetAffinity(xTaskGetCurrentTaskHandle()));

  while (true)
  {
    // start next timer run:
    tid++;
    if (tid % 1000 == 0)
      ESP_LOGI("app_task", "checkpoint tid=%u", tid);
    while (xTimerChangePeriod(timer, 1, portMAX_DELAY) != pdPASS);
    vTimerSetTimerID(timer, (void *) tid);
    while (xTimerStart(timer, portMAX_DELAY) != pdPASS);
    
    // wait for msg:
    if (xQueueReceive(queue, (void *) &msg, 100) == pdTRUE)
    {
      if (msg != tid)
        ESP_LOGE("app_task", "tid mismatch: received %u, expected %u", msg, tid);
    }
    else
    {
      ESP_LOGW("app_task", "queue rx timeout -- timer callback not issued!");
    }
  }
}


void app_main()
{
  // Print chip information
  esp_chip_info_t chip_info;
  esp_chip_info(&chip_info);
  printf("This is ESP32 chip with %d CPU cores, WiFi%s%s, ",
          chip_info.cores,
          (chip_info.features & CHIP_FEATURE_BT) ? "/BT" : "",
          (chip_info.features & CHIP_FEATURE_BLE) ? "/BLE" : "");
  printf("silicon revision %d, ", chip_info.revision);
  printf("%dMB %s flash\n", spi_flash_get_chip_size() / (1024 * 1024),
          (chip_info.features & CHIP_FEATURE_EMB_FLASH) ? "embedded" : "external");

  // Start wifi:
  start_wifi();
  vTaskDelay(5000 / portTICK_PERIOD_MS);

  // Start HTTP fetcher:
  //xTaskCreate(&http_get_task, "http_get_task", 4096, NULL, 5, NULL);

  // Start CPU/MEM load:
  xTaskCreate(&cpuload_task, "cpuload_task", 4096, NULL, 5, NULL);


#ifdef LED_OUT
  gpio_set_direction(LED_OUT, GPIO_MODE_OUTPUT);
#endif

  // create queue:
  queue = xQueueCreate(20, sizeof(uint32_t));

  // create some periodic timers:
  xTimerStart(xTimerCreate("per_timer", 5, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 6, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 7, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 10, pdTRUE, NULL, dummy_callback), portMAX_DELAY);
  xTimerStart(xTimerCreate("per_timer", 25, pdTRUE, NULL, dummy_callback), portMAX_DELAY);

  // create single shot timer:
  timer = xTimerCreate("test_timer", 1, pdFALSE, NULL, timer_callback);

  // timer service is on core 0, start application task on core 1:
  xTaskCreatePinnedToCore(app_task, "app_task", 4096, NULL, 5, NULL, 1);
}
